/*
 * Modular extensions service and registration functions
 *
 * Author: JuanJo Ciarlante <jjo-ipsec@mendoza.gov.ar>
 * 
 * Version: 0.7.1
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 */
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h> /* printk() */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/types.h>
#include <linux/string.h>	/* memcmp() */
#include <linux/random.h>	/* get_random_bytes() */
#include <linux/errno.h>  /* error codes */
#ifdef SPINLOCK
# ifdef SPINLOCK_23
#  include <linux/spinlock.h> /* *lock* */
# else /* SPINLOCK_23 */
#  include <asm/spinlock.h> /* *lock* */
# endif /* SPINLOCK_23 */
#endif /* SPINLOCK */
#ifdef NET_21
# include <asm/uaccess.h>
# include <linux/in6.h>
# define proto_priv cb
#endif /* NET21 */
#include "ipsec_param.h"
#include <freeswan.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_tunnel.h"
#include "ipsec_rcv.h"
#if defined(CONFIG_IPSEC_ESP) || defined(CONFIG_IPSEC_AH)
# include "ipsec_ah.h"
#endif /* defined(CONFIG_IPSEC_ESP) || defined(CONFIG_IPSEC_AH) */
#ifdef CONFIG_IPSEC_ESP
# include "ipsec_esp.h"
#endif /* !CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_IPCOMP
# include "ipcomp.h"
#endif /* CONFIG_IPSEC_COMP */

#include <pfkeyv2.h>
#include <pfkey.h>

#include "ipsec_sa.h"
#include "ipsec_ext.h"

static rwlock_t ipsec_ext_lock = RW_LOCK_UNLOCKED;

static struct ipsec_ext_enc *ipsec_ext_array_enc[SADB_EALG_MAX];
static struct ipsec_ext_auth *ipsec_ext_array_auth[SADB_AALG_MAX];	/* XXX jjo: auth */
#if SADB_EALG_MAX < 255
#warning Compiling with limited ESP support ( SADB_EALG_MAX < 256 )
#endif

/* already protected by lock */
static void __ipsec_ext_usage_inc(struct ipsec_ext *ixt) {
	if (ixt->ixt_module)
		__MOD_INC_USE_COUNT(ixt->ixt_module);
	atomic_inc(&ixt->ixt_refcnt);
}
static void __ipsec_ext_usage_dec(struct ipsec_ext *ixt) {
	atomic_dec(&ixt->ixt_refcnt);
	if (ixt->ixt_module)
		__MOD_DEC_USE_COUNT(ixt->ixt_module);
}
/*
 * 	Read-only kernel,user context -> no bh disabling 
 */
static struct ipsec_ext_enc *ipsec_ext_getby_encalg(int enc_alg) {
	struct ipsec_ext_enc *ixt_e = NULL;
	if (enc_alg >= SADB_EALG_MAX)
		goto out;
	read_lock(&ipsec_ext_lock);
	ixt_e=ipsec_ext_array_enc[enc_alg];
	if (ixt_e!=NULL) 
		__ipsec_ext_usage_inc((struct ipsec_ext *)ixt_e);
	read_unlock(&ipsec_ext_lock);
out:
	return ixt_e;
}
static struct ipsec_ext_auth *ipsec_ext_getby_authalg(int auth_alg) {
	struct ipsec_ext_auth *ixt_a = NULL;
	if (auth_alg > SADB_AALG_MAX)
		goto out;
	read_lock(&ipsec_ext_lock);
	ixt_a=ipsec_ext_array_auth[auth_alg];
	if (ixt_a!=NULL) 
		__ipsec_ext_usage_inc((struct ipsec_ext *)ixt_a);
	read_unlock(&ipsec_ext_lock);
out:
	return ixt_a;
}
void ipsec_ext_put(struct ipsec_ext *ixt) {
	__ipsec_ext_usage_dec((struct ipsec_ext *)ixt);
}
/* XXX: jjo check interface (ret values) */
int ipsec_ext_esp_encrypt(struct ipsec_sa *sa_p, __u8 * idat, int ilen, const __u8 * iv, int encrypt) {
	int ret;
	struct ipsec_ext_enc *ixt_e=sa_p->ips_ext_enc;
	KLIPS_PRINT(debug_rcv||debug_tunnel,
		    "klips_debug:ipsec_ext_esp_encrypt: "
		    "entering with encalg=%d, ixt_e=%p\n",
		    sa_p->ips_encalg, ixt_e);
	if (!ixt_e) {
		KLIPS_PRINT(debug_rcv||debug_tunnel,
			    "klips_debug:ipsec_ext_esp_encrypt: "
			    "NULL ipsec_ext_enc object\n");
		return -1;
	}
	KLIPS_PRINT(debug_rcv||debug_tunnel,
		    "klips_debug:ipsec_ext_esp_encrypt: "
		    "calling cbc_encrypt encalg=%d "
		    "ips_key_e=%p idat=%p ilen=%d iv=%p, encrypt=%d\n",
			sa_p->ips_encalg, 
			sa_p->ips_key_e, idat, ilen, iv, encrypt);
	ret=ixt_e->ixt_e_cbc_encrypt(sa_p->ips_key_e, idat, idat, ilen, iv, encrypt);
	KLIPS_PRINT(debug_rcv||debug_tunnel,
		    "klips_debug:ipsec_ext_esp_encrypt: "
		    "returned ret=%d\n",
		    ret);
#if 0	/* not needed, ipsec_sa has full ipsec_ext object ptr */
	ipsec_ext_put(ixt_e);
#endif
	return ret;
}
/* called from pfkey_v2_parser.c:pfkey_ips_init() */
int ipsec_ext_key_create(struct ipsec_sa *sa_p) {
	int ret=-EINVAL;
	int keyminbits, keymaxbits;
	caddr_t ekp;
	struct ipsec_ext_enc *ixt_e=sa_p->ips_ext_enc;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:ipsec_ext_key_create: "
		    "entering with encalg=%d ixt_e=%p\n",
		    sa_p->ips_encalg, ixt_e);
	if (!ixt_e) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:ipsec_ext_key_create: "
			    "NULL ipsec_ext_enc object\n");
		return -EPROTO;
	}
	if (ixt_e->ixt_e_validate_key) {
		if ((ret=ixt_e->ixt_e_validate_key(sa_p)) < 0) {
			KLIPS_PRINT(debug_pfkey,
				"klips_debug:pfkey_ipsec_sa_init: "
				"error calling validate_key() "
				"key_bits_e=%d\n",
				sa_p->ips_key_bits_e);
			goto ixt_out;
		}
	} else {
		keyminbits=ixt_e->ixt_keyminbits;
		keymaxbits=ixt_e->ixt_keymaxbits;
		if(sa_p->ips_key_bits_e<keyminbits || 
				sa_p->ips_key_bits_e>keymaxbits) {
			KLIPS_PRINT(debug_pfkey,
				"klips_debug:ipsec_ext_key_create: "
				"incorrect encryption key size: %d bits -- "
				"must be between %d,%d bits\n" /*octets (bytes)\n"*/,
				sa_p->ips_key_bits_e, keyminbits, keymaxbits);
			ret=-EINVAL;
			goto ixt_out;
		}
	}
	/* save encryption key pointer */
	ekp = sa_p->ips_key_e;

	if((sa_p->ips_key_e = (caddr_t)
	    kmalloc((sa_p->ips_key_e_size = ixt_e->ixt_e_ctx_size),
		    GFP_ATOMIC)) == NULL) {
		ret=-ENOMEM;
		goto ixt_out;
	}
	/* zero-out key_e */
	memset(sa_p->ips_key_e, 0, sa_p->ips_key_e_size);

	/* I cast here to allow more decoupling in alg module */
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:ipsec_ext_key_create: about to call:"
			    "set_key(key_e=%p, ekp=%p, key_size=%d)\n",
			    (caddr_t)sa_p->ips_key_e, ekp, sa_p->ips_key_bits_e/8);
	ret = ixt_e->ixt_e_set_key((caddr_t)sa_p->ips_key_e, ekp, sa_p->ips_key_bits_e/8);
	/* paranoid */
	memset(ekp, 0, sa_p->ips_key_bits_e/8);
	kfree(ekp);
ixt_out:
#if 0	/* not needed, ipsec_sa has full ipsec_ext object ptr */
	ipsec_ext_put(ixt_e);
#endif
	return ret;
}
int ipsec_ext_auth_create(struct ipsec_sa *sa_p) {
	int ret=-EINVAL;
	struct ipsec_ext_auth *ixt_a=sa_p->ips_ext_auth;
	int keyminbits, keymaxbits;
	unsigned char *akp;
	unsigned int aks;
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:ipsec_ext_auth_create: "
		    "entering with authalg=%d ixt_a=%p\n",
		    sa_p->ips_authalg, ixt_a);
	if (!ixt_a) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:ipsec_ext_key_create: "
			    "NULL ipsec_ext_auth object\n");
		return -EPROTO;
	}
	keyminbits=ixt_a->ixt_keyminbits;
	keymaxbits=ixt_a->ixt_keymaxbits;
	if(sa_p->ips_key_bits_a<keyminbits || sa_p->ips_key_bits_a>keymaxbits) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:ipsec_ext_auth_init: incorrect auth"
			    "key size: %d bits -- must be between %d,%d bits\n"/*octets (bytes)\n"*/,
			    sa_p->ips_key_bits_a, keyminbits, keymaxbits);
		ret=-EINVAL;
		goto ixt_out;
	}
	/* save auth key pointer */
	sa_p->ips_auth_bits = ixt_a->ixt_a_keylen * 8; /* XXX XXX */
	akp = sa_p->ips_key_a;
	aks = sa_p->ips_key_a_size;

	/* will hold: 2 ctx and a blocksize buffer: kb */
	sa_p->ips_key_a_size = ixt_a->ixt_a_ctx_size;
	if((sa_p->ips_key_a = 
		(caddr_t) kmalloc(sa_p->ips_key_a_size, GFP_ATOMIC)) == NULL) {
		ret=-ENOMEM;
		goto ixt_out;
	}
	ixt_a->ixt_a_hmac_set_key(sa_p->ips_key_a, akp, sa_p->ips_key_bits_a/8); /* XXX XXX */
	ret=0;
	memset(akp, 0, aks);
	kfree(akp);
			
ixt_out:
#if 0	/* not needed, ipsec_sa has full ipsec_ext object ptr */
	ipsec_ext_put(ixt_e);
#endif
	return ret;
}
/* ip_plen = len - iphlen  */
int ipsec_ext_sa_esp_hash(const struct ipsec_sa *sa_p, const __u8 *espp, int len, __u8 *hash, int hashlen) {
	struct ipsec_ext_auth *ixt_a=sa_p->ips_ext_auth;
	if (!ixt_a) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:ipsec_sa_esp_hash: "
			    "NULL ipsec_ext_auth object\n");
		return -EPROTO;
	}
	KLIPS_PRINT(debug_tunnel|debug_rcv,
			"klips_debug:ipsec_sa_esp_hash: "
			"hashing %p (%d bytes) to %p (%d bytes)\n",
			espp, len,
			hash, hashlen);
	ixt_a->ixt_a_hmac_hash(sa_p->ips_key_a, 
			espp, len,
			hash, hashlen);
	return 0;
}

/*	Old gcc's will fail here 	*/
#define barf_out(fmt, args...)  do { printk(KERN_ERR __FUNCTION__ ": (%s) " fmt, ixt->ixt_name , ## args)\
	; goto out; } while(0)

/* validation */
static int check_enc(struct ipsec_ext_enc *ixt) {
	int ret=-EINVAL;
	if (ixt->ixt_alg_id==0 || ixt->ixt_alg_id > SADB_EALG_MAX)
		barf_out("invalid alg_id=%d >= %d\n", ixt->ixt_alg_id, SADB_EALG_MAX);
	if (ixt->ixt_blocksize==0) /*  || ixt->ixt_blocksize%2) need for ESP_NULL */
		barf_out(KERN_ERR "invalid blocksize=%d\n", ixt->ixt_blocksize);
	if (ixt->ixt_keyminbits==0 && ixt->ixt_keymaxbits==0 && ixt->ixt_e_keylen==0)
		goto zero_key_ok;
	if (ixt->ixt_keyminbits==0)
		barf_out(KERN_ERR "invalid keyminbits=%d\n", ixt->ixt_keyminbits);
	if (ixt->ixt_keymaxbits==0)
		barf_out(KERN_ERR "invalid keymaxbits=%d\n", ixt->ixt_keymaxbits);
	if (ixt->ixt_e_keylen==0)
		barf_out(KERN_ERR "invalid keysize=%d\n", ixt->ixt_e_keylen);
zero_key_ok:
	if (ixt->ixt_e_ctx_size==0)
		barf_out(KERN_ERR "invalid key_e_size=%d\n", ixt->ixt_e_ctx_size);
	if (ixt->ixt_e_cbc_encrypt==NULL)
		barf_out(KERN_ERR "e_cbc_encrypt() must be not NULL\n");
	ret=0;
out:
	return ret;
}
static int check_auth(struct ipsec_ext_auth *ixt) {
	int ret=-EINVAL;
	if (ixt->ixt_alg_id==0 || ixt->ixt_alg_id > SADB_AALG_MAX)
		barf_out("invalid alg_id=%d > %d (SADB_AALG_MAX)\n", ixt->ixt_alg_id, SADB_AALG_MAX);
	if (ixt->ixt_blocksize==0 || ixt->ixt_blocksize%2)
		barf_out(KERN_ERR "invalid blocksize=%d\n", ixt->ixt_blocksize);
	if (ixt->ixt_blocksize>AH_BLKLEN_MAX)
		barf_out(KERN_ERR "sorry blocksize=%d > %d. "
			"Please increase AH_BLKLEN_MAX and recompile\n", 
			ixt->ixt_blocksize,
			AH_BLKLEN_MAX);
	if (ixt->ixt_keyminbits==0 && ixt->ixt_keymaxbits==0 && ixt->ixt_a_keylen==0)
		goto zero_key_ok;
	if (ixt->ixt_keyminbits==0)
		barf_out(KERN_ERR "invalid keyminbits=%d\n", ixt->ixt_keyminbits);
	if (ixt->ixt_keymaxbits==0)
		barf_out(KERN_ERR "invalid keymaxbits=%d\n", ixt->ixt_keymaxbits);
	if (ixt->ixt_keymaxbits!=ixt->ixt_keyminbits)
		barf_out(KERN_ERR "keymaxbits must equal keyminbits (not sure).\n");
	if (ixt->ixt_a_keylen==0)
		barf_out(KERN_ERR "invalid keysize=%d\n", ixt->ixt_a_keylen);
zero_key_ok:
	if (ixt->ixt_a_ctx_size==0)
		barf_out(KERN_ERR "invalid a_ctx_size=%d\n", ixt->ixt_a_ctx_size);
	if (ixt->ixt_a_hmac_set_key==NULL)
		barf_out(KERN_ERR "a_hmac_set_key() must be not NULL\n");
	if (ixt->ixt_a_hmac_hash==NULL)
		barf_out(KERN_ERR "a_hmac_hash() must be not NULL\n");
	ret=0;
out:
	return ret;
}
/* Generic registration entry point */
int register_ipsec_ext(struct ipsec_ext *ixt) {
	int ret=-EINVAL;
	struct ipsec_ext * const *ixt_p=NULL;
	/*	Validation 	*/
	if (ixt==NULL)
		barf_out("NULL ipsec_ext object passed\n");
	if ((ixt->ixt_version&0xffffff00) != (IPSEC_EXT_VERSION&0xffffff00))
		barf_out("incorrect version: %d.%d.%d-%d, "
			"must be %d.%d.%d[-%d]\n",
				IPSEC_EXT_VERSION_QUAD(ixt->ixt_version), 
				IPSEC_EXT_VERSION_QUAD(IPSEC_EXT_VERSION));
	switch(ixt->ixt_ext_type) {
		case IPSEC_EXT_TYPE_AUTH:
			if ((ret=check_auth((struct ipsec_ext_auth *)ixt)<0))
				goto out;
			ixt_p=(struct ipsec_ext **)
				&ipsec_ext_array_auth[ixt->ixt_alg_id];
			break;
		case IPSEC_EXT_TYPE_ENCRYPT: 
			if ((ret=check_enc((struct ipsec_ext_enc *)ixt)<0))
				goto out;
			ixt_p=(struct ipsec_ext **)
				&ipsec_ext_array_enc[ixt->ixt_alg_id];
			break;
		default:
			barf_out("ext_type=%d not supported", ixt->ixt_ext_type);
	}
	ret = -EINVAL;

	write_lock_bh(&ipsec_ext_lock);
	if (*ixt_p!=NULL) {
		write_unlock_bh(&ipsec_ext_lock);
		barf_out(KERN_WARNING "ipsec_ext for alg_id=%d already exist."
				"Not loaded (ret=%d).\n",
				ixt->ixt_alg_id, ret=-EEXIST);
	}
	*(struct ipsec_ext **)ixt_p=ixt;
	write_unlock_bh(&ipsec_ext_lock);

	if (ixt->ixt_ext_type == IPSEC_EXT_TYPE_ENCRYPT && ixt->ixt_ivlen == 0)
		ixt->ixt_ivlen = ixt->ixt_blocksize*8;

	ret = pfkey_list_insert_supported((struct supported *)&ixt->ixt_support, &(pfkey_supported_list[SADB_SATYPE_ESP]));
	if (ret==0) {
		ixt->ixt_state |= IPSEC_EXT_ST_SUPP;
		/*	send register event to userspace	*/
		pfkey_register_reply(SADB_SATYPE_ESP, NULL);
	} else
		printk(KERN_ERR "pfkey_list_insert_supported returned %d. "
				"Loading anyway.\n", ret);
	return 0;
out:
	return ret;
}
int unregister_ipsec_ext(struct ipsec_ext *ixt) {
	int ret= -EINVAL;
	struct ipsec_ext * const *ixt_p=NULL;
	switch(ixt->ixt_ext_type) {
		case IPSEC_EXT_TYPE_AUTH:
			ixt_p=(struct ipsec_ext **)
				&ipsec_ext_array_auth[ixt->ixt_alg_id];
			break;
		case IPSEC_EXT_TYPE_ENCRYPT: 
			ixt_p=(struct ipsec_ext **)
				&ipsec_ext_array_enc[ixt->ixt_alg_id];
			break;
	}
	if (ixt_p==NULL)
		barf_out("ERROR in logic (\"%s\"): ixt_p=NULL\n", 
				ixt->ixt_name);
	write_lock_bh(&ipsec_ext_lock);
	if (*ixt_p != ixt)  {
		write_unlock_bh(&ipsec_ext_lock);
		barf_out("array element #%d doesn't point to me\n",
				ixt->ixt_alg_id);
	}
	/*	This won't OOps because it points to static array 'slot' -jj */
	*(struct ipsec_ext **)ixt_p=NULL;

	if (ixt->ixt_state&IPSEC_EXT_ST_SUPP) {
		ixt->ixt_state &= ~IPSEC_EXT_ST_SUPP;
		pfkey_list_remove_supported((struct supported *)&ixt->ixt_support, &(pfkey_supported_list[SADB_SATYPE_ESP]));
		/*	send register event to userspace	*/
		pfkey_register_reply(SADB_SATYPE_ESP, NULL);
	}

	write_unlock_bh(&ipsec_ext_lock);
	ret=0;
out:
	return ret;
}
/*
 * 	Must be called from user context
 */
static int ipsec_ext_test_encrypt(int enc_alg, int test) {
	int ret;
	caddr_t buf;
	int iv_size, keysize, key_e_size;
	struct ipsec_ext_enc *ixt_e;
	#define BUFSZ	1024
	#define MARGIN	0
	#define test_enc   (buf+MARGIN)
	#define test_dec   (test_enc+BUFSZ+MARGIN)
	#define test_tmp   (test_dec+BUFSZ+MARGIN)
	#define test_key_e (test_tmp+BUFSZ+MARGIN)
	#define test_iv    (test_key_e+key_e_size+MARGIN)
	#define test_key   (test_iv+iv_size+MARGIN)
	#define test_size  (BUFSZ*3+key_e_size+iv_size+keysize+MARGIN*7)
	ixt_e=ipsec_ext_getby_encalg(enc_alg);
	if (ixt_e==NULL) {
		KLIPS_PRINT(1, 
			    "klips_debug:" __FUNCTION__ ": "
			    "encalg=%d object not found\n",
			    enc_alg);
		ret=-EINVAL;
		goto out;
	}
	iv_size=ixt_e->ixt_blocksize;
	key_e_size=ixt_e->ixt_e_ctx_size;
	keysize=ixt_e->ixt_e_keylen;
	KLIPS_PRINT(1, 
		    "klips_debug:" __FUNCTION__ ": "
		    "enc_alg=%d blocksize=%d key_e_size=%d keysize=%d\n",
		    enc_alg, iv_size, key_e_size, keysize);
	if ((buf=kmalloc (test_size, GFP_KERNEL)) == NULL) {
		ret= -ENOMEM;
		goto out;
	}
	get_random_bytes(test_key, keysize);
	get_random_bytes(test_iv, iv_size);
	ixt_e->ixt_e_set_key(test_key_e, test_key, keysize);
	get_random_bytes(test_enc, BUFSZ);
	memcpy(test_tmp, test_enc, BUFSZ);
	ret=ixt_e->ixt_e_cbc_encrypt(test_key_e, test_enc, test_enc, BUFSZ, test_iv, 1);
	printk(KERN_INFO
		    "klips_info:" __FUNCTION__ ": "
		    "cbc_encrypt=1 ret=%d\n", 
		    	ret);
	ret=memcmp(test_enc, test_tmp, BUFSZ);
	printk(KERN_INFO
		    "klips_info:" __FUNCTION__ ": "
		    "memcmp(enc, tmp) ret=%d: %s\n", ret,
			ret!=0? "OK. (encr->DIFFers)" : "FAIL! (encr->SAME)" );
	memcpy(test_dec, test_enc, BUFSZ);
	ret=ixt_e->ixt_e_cbc_encrypt(test_key_e, test_dec, test_dec, BUFSZ, test_iv, 0);
	printk(KERN_INFO
		    "klips_info:" __FUNCTION__ ": "
		    "cbc_encrypt=0 ret=%d\n", ret);
	ret=memcmp(test_dec, test_tmp, BUFSZ);
	printk(KERN_INFO
		    "klips_info:" __FUNCTION__ ": "
		    "memcmp(dec,tmp) ret=%d: %s\n", ret,
			ret==0? "OK. (encr->decr->SAME)" : "FAIL! (encr->decr->DIFFers)" );
	{
		/*	Shamelessly taken from drivers/md sources  O:)  */
		unsigned long now;
		int i, count, max=0;
		int encrypt, speed;
		for (encrypt=0; encrypt <2;encrypt ++) {
			for (i = 0; i < 5; i++) {
				now = jiffies;
				count = 0;
				while (jiffies == now) {
					mb();
					ixt_e->ixt_e_cbc_encrypt(test_key_e, test_tmp, test_tmp, BUFSZ, test_iv, encrypt);
					mb();
					count++;
					mb();
				}
				if (count > max)
					max = count;
			}
			speed = max * (HZ * BUFSZ / 1024);
			printk(KERN_INFO
				    "klips_info:" __FUNCTION__ ": "
				    "%s speed=%d KB/s\n", 
				    encrypt? "encrypt": "decrypt", speed);
		}
	}
	kfree(buf);
	ipsec_ext_put((struct ipsec_ext *)ixt_e);
out:
	return ret;
	#undef test_enc  
	#undef test_dec  
	#undef test_tmp  
	#undef test_key_e
	#undef test_iv   
	#undef test_key  
	#undef test_size 
}
/*
 * 	Must be called from user context
 */
static int ipsec_ext_test_auth(int enc_alg, int test) {
	int ret;
	caddr_t buf;
	int blocksize, keysize, key_a_size;
	struct ipsec_ext_auth *ixt_a;
	#define BUFSZ	1024
	#define MARGIN	0
	#define test_auth  (buf+MARGIN)
	#define test_key_a (test_auth+BUFSZ+MARGIN)
	#define test_key   (test_key_a+key_a_size+MARGIN)
	#define test_hash  (test_key+keysize+MARGIN)
	#define test_size  (BUFSZ+key_a_size+keysize+AHHMAC_HASHLEN+MARGIN*4)
	ixt_a=ipsec_ext_getby_authalg(enc_alg);
	if (ixt_a==NULL) {
		KLIPS_PRINT(1, 
			    "klips_debug:" __FUNCTION__ ": "
			    "encalg=%d object not found\n",
			    enc_alg);
		ret=-EINVAL;
		goto out;
	}
	blocksize=ixt_a->ixt_blocksize;
	key_a_size=ixt_a->ixt_a_ctx_size;
	keysize=ixt_a->ixt_a_keylen;
	KLIPS_PRINT(1, 
		    "klips_debug:" __FUNCTION__ ": "
		    "enc_alg=%d blocksize=%d key_a_size=%d keysize=%d\n",
		    enc_alg, blocksize, key_a_size, keysize);
	if ((buf=kmalloc (test_size, GFP_KERNEL)) == NULL) {
		ret= -ENOMEM;
		goto out;
	}
	get_random_bytes(test_key, keysize);
	ixt_a->ixt_a_hmac_set_key(test_key_a, test_key, keysize);
	get_random_bytes(test_auth, BUFSZ);
	ret=ixt_a->ixt_a_hmac_hash(test_key_a, test_auth, BUFSZ, test_hash, AHHMAC_HASHLEN);
	printk(KERN_INFO
		    "klips_info:" __FUNCTION__ ": "
		    "ret=%d\n", ret);
	{
		/*	Shamelessly taken from drivers/md sources  O:)  */
		unsigned long now;
		int i, count, max=0;
		int speed;
		for (i = 0; i < 5; i++) {
			now = jiffies;
			count = 0;
			while (jiffies == now) {
				mb();
				ixt_a->ixt_a_hmac_hash(test_key_a, test_auth, BUFSZ, test_hash, AHHMAC_HASHLEN);
				mb();
				count++;
				mb();
			}
			if (count > max)
				max = count;
		}
		speed = max * (HZ * BUFSZ / 1024);
		printk(KERN_INFO
				"klips_info:" __FUNCTION__ ": "
				"speed=%d KB/s\n", speed);
	}
	kfree(buf);
	ipsec_ext_put((struct ipsec_ext *)ixt_a);
out:
	return ret;
	#undef test_auth 
	#undef test_key_a
	#undef test_key  
	#undef test_hash 
	#undef test_size 
}
int ipsec_ext_test(unsigned ext_type, unsigned alg_id, int test) {
	switch(ext_type) {
		case IPSEC_EXT_TYPE_ENCRYPT:
			return ipsec_ext_test_encrypt(alg_id, test);
			break;
		case IPSEC_EXT_TYPE_AUTH:
			return ipsec_ext_test_auth(alg_id, test);
			break;
	}
	printk(KERN_ERR "klips_info: ipsec_ext_test() called incorrectly: "
			"ext_type=%d alg_id=%d\n",
			ext_type, alg_id);
	return -EINVAL;
}
int ipsec_ext_init(void) {
	KLIPS_PRINT(1, "klips_info:ipsec_ext_init: "
			"KLIPS ext v=%d.%d.%d-%d (EALG_MAX=%d, AALG_MAX=%d)\n",
			IPSEC_EXT_VERSION_QUAD(IPSEC_EXT_VERSION),
			SADB_EALG_MAX, SADB_AALG_MAX);
	write_lock_bh(&ipsec_ext_lock);
	memset(ipsec_ext_array_enc, 0, sizeof(ipsec_ext_array_enc)); 
	memset(ipsec_ext_array_auth, 0, sizeof(ipsec_ext_array_auth)); 
	write_unlock_bh(&ipsec_ext_lock);
	/*	Initialize static algos 	*/
	KLIPS_PRINT(1, "klips_info:ipsec_ext_init: "
		"calling ipsec_ext_static_init()\n");
	ipsec_ext_static_init();
	return 0;
}

/**************************************
 *
 * 	ipsec_ext sa interface
 *
 **************************************/

/*	
 *	Called from pluto -> pfkey_v2_parser.c:pfkey_ipsec_sa_init()	
 */
int ipsec_ext_sa_init(struct ipsec_sa *sa_p) {
	struct ipsec_ext_enc *ixt_e;
	struct ipsec_ext_auth *ixt_a;

	/*	Only ESP for now ... */
	if (sa_p->ips_said.proto != IPPROTO_ESP)
		return -EPROTONOSUPPORT;
	KLIPS_PRINT(debug_pfkey, "klips_debug: " __FUNCTION__ "() :"
			"entering for encalg=%d, authalg=%d\n",
			    sa_p->ips_encalg, sa_p->ips_authalg);
	if ((ixt_e=ipsec_ext_getby_encalg(sa_p->ips_encalg))) {
		KLIPS_PRINT(debug_pfkey,
		    "klips_debug: " __FUNCTION__ "() :"
		    "found ipsec_ext (ixt_e=%p) for encalg=%d\n",
		    ixt_e, sa_p->ips_encalg);
		sa_p->ips_ext_enc=ixt_e;
	}
	if ((ixt_a=ipsec_ext_getby_authalg(sa_p->ips_authalg))) {
		KLIPS_PRINT(debug_pfkey,
		    "klips_debug: " __FUNCTION__ "() :"
		    "found ipsec_ext (ixt_a=%p) for auth=%d\n",
		    ixt_a, sa_p->ips_authalg);
		sa_p->ips_ext_auth=ixt_a;
	}
	return 0;
}

/*	
 *	Called from pluto -> ipsec_sa.c:ipsec_sa_delchain()
 */
int ipsec_ext_sa_wipe(struct ipsec_sa *sa_p) {
	struct ipsec_ext *ixt;
	if ((ixt=(struct ipsec_ext *)sa_p->ips_ext_enc)) {
		KLIPS_PRINT(debug_pfkey, "klips_debug: " __FUNCTION__ "() :"
				"unlinking for encalg=%d\n",
				ixt->ixt_alg_id);
		ipsec_ext_put(ixt);
	}
	if ((ixt=(struct ipsec_ext *)sa_p->ips_ext_auth)) {
		KLIPS_PRINT(debug_pfkey, "klips_debug: " __FUNCTION__ "() :"
				"unlinking for authalg=%d\n",
				ixt->ixt_alg_id);
		ipsec_ext_put(ixt);
	}
	return 0;
}
/*
 * 	As the author of this module, I ONLY ALLOW using it from
 * 	GPL (or same LICENSE TERMS as kernel source) modules.
 *
 * 	And NO... I don't want even closed-source hardware ciphers 
 * 	to use this interface.
 * 	--Juan Jose Ciarlante
 */

/*	
 *	These symbols can only be used from GPL modules	
 *	for now, I'm disabling this because it creates false
 *	symbol problems for old modutils.
 */

/* #ifndef EXPORT_SYMBOL_GPL */
#undef EXPORT_SYMBOL_GPL
#define EXPORT_SYMBOL_GPL EXPORT_SYMBOL
/* #endif */
EXPORT_SYMBOL_GPL(register_ipsec_ext);
EXPORT_SYMBOL_GPL(unregister_ipsec_ext);
EXPORT_SYMBOL_GPL(ipsec_ext_test);
